/*

*/



/**
   {@code List<E>} represents a generic list of values that can grow and shrink.
   <p>
   The public interface of this list ADT has the following methods.
   <p>
      int     size()                       <br>
      void    add(int index, E e)          <br>
      E       remove(int index)            <br>
      boolean remove(E e)                  <br>
      void    set(int index, E e)          <br>
      E       get(int index)               <br>
      void    clear()                      <br>
      boolean isEmpty()                    <br>
      boolean contains(E e)                <br>
      int     indexOf(E e)                 <br>
      int     lastIndexOf(E e)             <br>
*/
public abstract class List<E>
{
   /**
      Returns the number of items currently in this {@code List}.

      @return the number of items in this {@code List}
   */
   public abstract int size();


   /**
      Returns true if this {@code List} does not conatin any items.

      @return true if there are no items in this {@code List}
   */
   public boolean isEmpty()
   {
      return 0 == size();
   }


   /**
      Return the item at the given {@code index} in this {@code List}.

      @param index  location in this {@code List} of the item to retrieve
      @return a reference to the item in this {@code List} at location {@code index}
   */
   public abstract E get(int index);


   /**
      Set the value of the {@code List} item at the given {@code index}.

      @param index  location in this {@code List} of the item to replace
      @param item   reference to the item to place in this {@code List}
   */
   public abstract void set(int index, E item);


   /**
      Insert the item at the given {@code index} in this {@code List}.
      The {@code List} size increases by one.
      <p>
      For this method, the index is valid if 0 <= index <= size().
      <p>
      If the index is equal to the size of the list, that means that we are
      adding an elelment to the end of the list. In particular, if the list
      is empty, we can add an item at index 0, even though 0 is not (yet)
      a valid index into the list.

      @param index  location in this {@code List} for the inserted {@code item}
      @param item   reference to the item to be added to this {@code List}
   */
   public abstract void add(int index, E item);


   /**
      Remove the item at the given {@code index} in this {@code List}.
      The {@code List} size decreases by one.

      @param index  location in this {@code List} of the item to remove
      @return a reference to the item that was removed
   */
   public abstract E remove(int index);


   /**
      Remove the first occurrence of the given item from this {@code List}.
      Return {@code true} if the {@code item} is found, otherwise return {@code false}.

      @param item  reference to the item to search for and remove from this {@code List}
      @return true if the {@code item} was found and removed, false otherwise
   */
   public boolean remove(E item)
   {
      int index = this.indexOf(item);

      if (-1 == index)
      {
         return false;
      }

      this.remove(index);

      return true;
   }


   /**
      Remove all the items in this {@code List}.
      The {@code List} size becomes zero.
   */
   public abstract void clear();


   /**
      Search this {@code List} for the given {@code item}.
      Return {@code true} if the {@code item} is found.

      @param item  reference to the item to search for in this {@code List}
      @return true if the {@code item} was found in this {@code List}
   */
   public boolean contains(E item)
   {
      return -1 != this.indexOf(item);
   }


   /**
      Search this {@code List} for the given {@code item}.
      Return the index of the first occurrence of the
      {@code item} if it is found and return {@code -1}
      if the item is not found.

      @param item  reference to the item to search for in this {@code List}
      @return the first index of {@code item} if it is found, -1 otherwise
   */
   public abstract int indexOf(E item);


   /**
      Search this {@code List} for the given {@code item}.
      Return the index of the last occurrence of the
      {@code item} if it is found and return {@code -1}
      if the item is not found.

      @param item  reference to the item to search for in this {@code List}
      @return the last index of {@code item} if it is found, -1 otherwise
   */
   public abstract int lastIndexOf(E item);


   /**
      Return a {@link String} representation of this {@code List}.

      @return {@link String} representation of this {@code List}
   */
   @Override
   public String toString()
   {
      String result = "[";
      for (int i = 0; i < this.size(); ++i)
      {
         result += this.get(i);
         if (i < this.size() - 1)
            result += ", ";
      }
      result += "]";
      return result;
   }
}
